/*
Copyright 2015-2016, Manuel Bichler
WWW: <http://dbai.tuwien.ac.at/proj/lpopt/>

This file is part of lpopt.

lpopt is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

lpopt is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lpopt.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <algorithm>
#include <memory>

#include "safetyCharacterization.hpp"

using namespace std;

namespace domain {

  bool rule_cmp::operator() (SafetyRule * lhs, SafetyRule * rhs) const
  {
    return (*lhs) < (*rhs);
  }

  SafetyRule::SafetyRule(varset* dependees, const Variable* depender) : dependees(dependees), depender(depender) {}
  
  SafetyRule::~SafetyRule()
  {
    delete dependees;
  }

  const Variable* SafetyRule::perform(const varset& safes) const
  {
    if (std::includes(safes.begin(), safes.end(), dependees->begin(), dependees->end(), var_cmp()))
      return depender;
    else
      return NULL;
  }
  
  const Variable& SafetyRule::getDepender() const
  {
    return *depender;
  }
  const varset& SafetyRule::getDependees() const
  {
    return *dependees;
  }

  SafetyRule::SafetyRule(SafetyRule const &that) : dependees(new varset(*that.dependees)), depender(that.depender) {}
  
  SafetyRule::SafetyRule(SafetyRule&& other) : SafetyRule()
  {
    swap(*this, other);
  }

  SafetyRule* SafetyRule::clone() const {
    return new SafetyRule(*this);
  }

  void swap(SafetyRule& first, SafetyRule& second)
  {
    using std::swap;
    swap(first.dependees, second.dependees);
    swap(first.depender, second.depender);
  }
  
  SafetyRule& SafetyRule::operator=(SafetyRule that)
  {
    swap(*this, that);
    return *this;
  }

  inline bool operator==(const SafetyRule& lhs, const SafetyRule& rhs){
    if (*lhs.depender != *rhs.depender) return false;
    if (lhs.dependees->size() != rhs.dependees->size()) return false;
    if (std::equal(lhs.dependees->begin(), lhs.dependees->end(), rhs.dependees->begin(), var_cmp())) return true;
    return false;
  }
  inline bool operator!=(const SafetyRule& lhs, const SafetyRule& rhs){return !operator==(lhs,rhs);}
  inline bool operator< (const SafetyRule& lhs, const SafetyRule& rhs){
    if (*lhs.depender >= *rhs.depender) return false;
    if (lhs.dependees->size() < rhs.dependees->size()) return true;
    if (lhs.dependees->size() > rhs.dependees->size()) return false;
    varset::iterator rit = rhs.dependees->begin();
    for (varset::iterator lit = lhs.dependees->begin(); lit!=lhs.dependees->end(); ++lit)
    {
      if (**lit < **rit) return true;
      if (**lit > **rit) return false;
    }
    return false;
  }
  inline bool operator> (const SafetyRule& lhs, const SafetyRule& rhs){return  operator< (rhs,lhs);}
  inline bool operator<=(const SafetyRule& lhs, const SafetyRule& rhs){return !operator> (lhs,rhs);}
  inline bool operator>=(const SafetyRule& lhs, const SafetyRule& rhs){return !operator< (lhs,rhs);}

  bool SafetyRule::includes(const SafetyRule &other) const
  {
    if (*depender != *(other.depender))
      return false;
    if (std::includes(other.dependees->begin(), other.dependees->end(), this->dependees->begin(), this->dependees->end(), var_cmp()))
      return true;
    return false;
  }
  void SafetyRule::incorporate(const SafetyRule &other)
  {
    if (std::includes(this->dependees->begin(), this->dependees->end(), other.dependees->begin(), other.dependees->end(), var_cmp()))
      this->dependees->erase(other.depender);
  }
  void SafetyRule::removeDependees(const varset& vars)
  {
    varset::iterator depit = dependees->begin();
    varset::iterator varit = vars.begin();
    while(depit != dependees->end() && varit != vars.end())
    {
      if(var_cmp()(*depit, *varit))
	++depit;
      else if (var_cmp()(*varit, *depit))
	++varit;
      else { // equal
	depit = dependees->erase(depit);
	++varit;
      }
    }
  }

  

  SafetyCharacterization::SafetyCharacterization() : safes(new varset()), unsafes(new varset()) {}
  
  SafetyCharacterization::SafetyCharacterization(const Variable* var, bool safe) : SafetyCharacterization()
  {
    if(safe)
      safes->insert(var);
    else
      unsafes->insert(var);
  }

  SafetyCharacterization::SafetyCharacterization(varset* vars, bool safe)
  {
    if (safe)
    {
      safes = vars;
      unsafes = new varset();
    }
    else
    {
      unsafes = vars;
      safes = new varset();
    }
  }

  void SafetyCharacterization::addRule(SafetyRule *rule)
  {
    if (safes->find(&(rule->getDepender())) != safes->end())
    {
      delete rule;
      return; // variable already safe
    }
    const Variable *v = rule->perform(*safes);
    if (v!=NULL)
    {
      // the rule fires instantly! No need to safe it, just add the depender.
      delete rule;
      addSafe(v);
      return;
    }
    // not yet fireable. No need to change other sets, just insert.
    for (ruleset::iterator it = rules.begin(); it != rules.end(); ++it)
    {
      if ((*it)->includes(*rule))
      { // information already inside
	delete rule;
	return;
      }
      if (rule->includes(**it))
      { // old rule unneccessary
	delete *it;
	it = rules.erase(it);
	--it;
      }
      (*it)->incorporate(*rule);
    }
    rules.insert(rule);
  }

  bool SafetyCharacterization::addSafe(const Variable* var)
  {
    bool b = safes->insert(var).second;
    if (b) // it was not yet inside
      rectifyAfterAddSafes();
    return b;
  }
  
  bool SafetyCharacterization::addUnsafe(const Variable* var)
  {
    if (safes->find(var) != safes->end()) return false; // this variable is safe. Don't add.
    return unsafes->insert(var).second;
  }
  
  void SafetyCharacterization::addSafe(const varset& vars)
  {
    safes->insert(vars.begin(), vars.end());
    rectifyAfterAddSafes();
  }
  
  void SafetyCharacterization::addUnsafe(const varset& vars)
  {
    // don't add already safe ones:
    vector<const Variable*> toAdd(vars.size());
    vector<const Variable*>::iterator it = std::set_difference(vars.begin(), vars.end(), safes->begin(), safes->end(), toAdd.begin(), var_cmp());
    toAdd.resize(it - toAdd.begin());

    unsafes->insert(toAdd.begin(), toAdd.end());
  }
  
  bool SafetyCharacterization::add(const Variable* var, bool safe)
  {
    if (safe) return addSafe(var);
    else return addUnsafe(var);
  }
  
  void SafetyCharacterization::add(const varset& vars, bool safe)
  {
    if (safe) addSafe(vars);
    else addUnsafe(vars);
  }
  
  void SafetyCharacterization::rectifyAfterAddSafes()
  {
    // let rules trigger:
    bool changed = true;
    while(changed)
    {
      changed = false;
      for (ruleset::iterator it = rules.begin(); it!=rules.end(); ++it)
      {
	const Variable* v = (*it)->perform(*safes);
	if (v!=NULL)
	{
	  safes->insert(v);
	  delete (*it);
	  it = rules.erase(it);
// 	  --it;
	  changed = true;
	  break; // TODO: actually, I would rather like the version --it more, but getting segfault in operator--().
	}
	else
	  (*it)->removeDependees(*safes);
      }
    }
    
    // new safe vars are not unsafe:
    for (varset::iterator it = safes->begin(); it != safes->end(); ++it)
      unsafes->erase(*it); // experimental result (after hours of debugging): unsafes->erase(safes->begin(), safes->end()) erases from safes, not from unsafes.
    
    //delete rules that entail safe_vars (that might be others than those that triggered in the loop above):
    if (safes->empty()) return;
    ruleset::iterator rit = rules.begin();
    while(rit!=rules.end())
    {
      if (safes->find(&(*rit)->getDepender()) != safes->end())
      {
	delete (*rit);
	rit = rules.erase(rit); // rit now points to next rule
      }
      else
	++rit;
    }
  }

  bool SafetyCharacterization::isSafe(const Variable* v) const
  {
    return safes->find(v) != safes->end();
  }
  
  bool SafetyCharacterization::isSafe(const Variable* v, const varset& safes) const
  {
    if (isSafe(v))
      return true;
    unique_ptr<SafetyCharacterization> cp(clone());
    cp->addSafe(safes);
    return cp->isSafe(v);
  }

  const varset& SafetyCharacterization::safeVars() const
  {
    return *safes;
  }

  const varset& SafetyCharacterization::unsafeVars() const
  {
    return *unsafes;
  }
  
  varset* SafetyCharacterization::nonSafeable() const
  {
    varset* ret = new varset(*unsafes);
    for (ruleset::iterator rit = rules.begin(); rit!=rules.end(); ++rit)
      ret->erase(&((*rit)->getDepender()));
    return ret;
  }
  set<varset*>* SafetyCharacterization::savingCandidates(const Variable* v) const
  {
    set<varset*>* ret = new set<varset*>();
    if (isSafe(v))
    {
      ret->insert(new varset());
      return ret;
    }
    for (ruleset::iterator rit = rules.begin(); rit!=rules.end(); ++rit)
      if ((*rit)->getDepender() == *v)
	ret->insert(new varset((*rit)->getDependees()));
    return ret;
  }
  bool SafetyCharacterization::saveable(const Variable* v) const
  {
    for (ruleset::iterator rit = rules.begin(); rit!=rules.end(); ++rit)
      if ((*rit)->getDepender() == *v)
	return true;
    if (isSafe(v))
      return true;
    return false;
  }
  
  ruleset* SafetyCharacterization::candidateRules(const Variable& unsafe) const
  {
    ruleset* ret = new ruleset();
    for (ruleset::iterator rit = rules.begin(); rit!=rules.end(); ++rit)
      if ((*rit)->getDepender() == unsafe)
	ret->insert(*rit);
    return ret;
  }

  SafetyCharacterization* SafetyCharacterization::incorporate(const SafetyCharacterization& other)
  {
    // first, copy it all over:
    addUnsafe(*(other.unsafes));
    if (other.safes->empty() && other.rules.empty())
      return this; // no new information to trigger rules
    safes->insert(other.safes->begin(), other.safes->end());
    // rule copying complicated because sets hold pointers but we want to make deep copy
    vector<const SafetyRule*> toAdd(other.rules.size());
    vector<const SafetyRule*>::iterator it = std::set_difference(other.rules.begin(), other.rules.end(), rules.begin(), rules.end(), toAdd.begin(), rule_cmp());
    toAdd.resize(it - toAdd.begin());
    ruleset::iterator hint = rules.begin();
    for (it = toAdd.begin(); it != toAdd.end(); ++it)
    {
      if (safes->find(&(*it)->getDepender()) == safes->end()) // only add rule if depender not already safe
	hint = rules.insert(hint, (*it)->clone());
    }

    rectifyAfterAddSafes();
    
    return this;
  }
  
  SafetyCharacterization* SafetyCharacterization::combine(SafetyCharacterization* other)
  { // TODO: might be more efficient to explicitly implement instead of using incorporate
    incorporate(*other);
    delete other;
    return this;
  }
  
  SafetyCharacterization* SafetyCharacterization::equalsCombine(SafetyCharacterization* other)
  {
    /* the atom's safety definition set consists of the safety implications:
     * (left_safe union left_unsafe) implies (right_safe).
     * (right_safe union right_unsafe) implies (left_safe).
     * (left_unsafe) implies.
     * (left_safe) implies.
     * (right_unsafe) implies.
     * (right_safe) implies.
     * in each implication, right-side variables are omitted, if they also occur on the left side.
     * implications of form (set) implies (empty) are split up into implications (variable) implies (empty) for each variable in the set; They are considered unsafe.
     * vice versa: implications of the form (empty) implies (set) are split up into implications (empty) implies (variable) for each variable in the set; they are considered safe.
     * if a variable occurs safe, it does not count as unsafe: implications of form (varaible) implies (empty) are omitted if there is an implication (empty) implies (variable).
     */
    if (this->safes->empty())
    {
      if (other->safes->empty())
	forceAddUnsafe(*(other->unsafes)); // this has both leftUnsafe and rightUnsafe as unsafes
      else
      {
	for (varset::iterator it = other->safes->begin(); it != other->safes->end(); ++it)
	  rules.insert(new SafetyRule( new varset(*unsafes), *it));
	forceAddUnsafe(*(other->unsafes));
	forceAddUnsafe(*(other->safes)); // this has leftUnsafe, rightUnsafe and rightSafe as unsafes
	rectifyAfterAddSafes();
      }
      delete other;
      return this;
    } else if (other->safes->empty())
    {
      for (varset::iterator it = safes->begin(); it != safes->end(); ++it)
	rules.insert(new SafetyRule( it == --safes->end() ? other->unsafes : new varset(*(other->unsafes)), *it));
      forceAddUnsafe(*(other->unsafes));
      forceAddUnsafe(*safes); // this has leftUnsafe, rightUnsafe and leftSafe as unsafes
      safes->clear();
      other->unsafes = NULL; // we have used the object in last rule
      delete other;
      rectifyAfterAddSafes();
      return this; 
    }
    
    // little more complicated: both this and other have safes
    varset* leftAll = new varset(*unsafes);
    leftAll->insert(safes->begin(), safes->end());
    
    varset* rightAll = other->unsafes;
    rightAll->insert(other->safes->begin(), other->safes->end());
    other->unsafes = NULL;

    forceAddUnsafe(*safes);
    forceAddUnsafe(*rightAll); // this has leftUnsafe, rightUnsafe, leftSafe and rightSafe as unsafes
    
    for (varset::iterator it = other->safes->begin(); it != other->safes->end(); ++it)
      rules.insert(new SafetyRule( it == --other->safes->end() ? leftAll : new varset(*leftAll), *it));
    for (varset::iterator it = safes->begin(); it != safes->end(); ++it)
      rules.insert(new SafetyRule( it == --safes->end() ? rightAll : new varset(*rightAll), *it));
    
    safes->clear();
    delete other;
    rectifyAfterAddSafes();
    return this;
  }
  
  SafetyCharacterization* SafetyCharacterization::equalsIncorporate(const SafetyCharacterization& other)
  {
    if (this->safes->empty())
    {
      if (other.safes->empty())
	forceAddUnsafe(*(other.unsafes)); // this has both leftUnsafe and rightUnsafe as unsafes
      else
      {
	for (varset::iterator it = other.safes->begin(); it != other.safes->end(); ++it)
	  rules.insert(new SafetyRule( new varset(*unsafes), *it));
	forceAddUnsafe(*(other.unsafes));
	forceAddUnsafe(*(other.safes)); // this has leftUnsafe, rightUnsafe and rightSafe as unsafes
	rectifyAfterAddSafes();
      }
      return this;
    } else if (other.safes->empty())
    {
      for (varset::iterator it = safes->begin(); it != safes->end(); ++it)
	rules.insert(new SafetyRule(new varset(*(other.unsafes)), *it));
      forceAddUnsafe(*(other.unsafes));
      forceAddUnsafe(*safes); // this has leftUnsafe, rightUnsafe and leftSafe as unsafes
      safes->clear();
      rectifyAfterAddSafes();
      return this; 
    }
    
    //little more complicated: both this and other have safes
    varset* leftAll = new varset(*unsafes);
    leftAll->insert(safes->begin(), safes->end());
    
    varset* rightAll = new varset(*other.unsafes);
    rightAll->insert(other.safes->begin(), other.safes->end());

    forceAddUnsafe(*safes);
    forceAddUnsafe(*rightAll); // this has leftUnsafe, rightUnsafe, leftSafe and rightSafe as unsafes
    
    for (varset::iterator it = other.safes->begin(); it != other.safes->end(); ++it)
      rules.insert(new SafetyRule( it == --other.safes->end() ? leftAll : new varset(*leftAll), *it));
    for (varset::iterator it = safes->begin(); it != safes->end(); ++it)
      rules.insert(new SafetyRule( it == --safes->end() ? rightAll : new varset(*rightAll), *it));
    
    safes->clear();
    rectifyAfterAddSafes();
    return this;
  }

  bool SafetyCharacterization::isSimple() const {
    return rules.empty();
  }
  
  void SafetyCharacterization::forceAddUnsafe(const varset& vars)
  {
    unsafes->insert(vars.begin(), vars.end());
  }

  
  SafetyCharacterization::~SafetyCharacterization()
  {
    delete safes;
    delete unsafes;
    for (ruleset::iterator it = rules.begin(); it != rules.end(); ++it)
      delete *it;
  }

  SafetyCharacterization::SafetyCharacterization(SafetyCharacterization const &that) : safes(new varset(*that.safes)), unsafes(new varset(*that.unsafes)), rules()
  {
    for (ruleset::iterator it = that.rules.begin(); it != that.rules.end(); ++it)
      rules.insert(rules.end(), (*it)->clone()); // deep copy
  }
  
  SafetyCharacterization::SafetyCharacterization(SafetyCharacterization&& other) : SafetyCharacterization()
  {
    swap(*this, other);
  }

  SafetyCharacterization* SafetyCharacterization::clone() const {
    return new SafetyCharacterization(*this);
  }

  void swap(SafetyCharacterization& first, SafetyCharacterization& second)
  {
    using std::swap;
    swap(first.safes, second.safes);
    swap(first.unsafes, second.unsafes);
    swap(first.rules, second.rules);
  }
  
  SafetyCharacterization& SafetyCharacterization::operator=(SafetyCharacterization that)
  {
    swap(*this, that);
    return *this;
  }

  
} // namespace domain
